@@ -24,6 +24,7 @@
 #include <QToolBar>
 #include <QMainWindow>
 #include <QPushButton>
+#include <QCommandLinkButton>
 #include <QComboBox>
 #include <QLineEdit>
 #include <QProgressBar>
@@ -38,6 +39,7 @@
 #include <QDialog>
 #include <QWindow> // for positioning menus
 #include <QScreen> // for positioning menus
+#include <QStylePainter>
 
 namespace Kvantum
 {
@@ -183,10 +185,10 @@ void Style::drawBg(QPainter *p, const QWidget *widget) const
   int dw = sspec.incrementW ? sspec.minW : qMax(sspec.minW - bgndRect.width(), 0);
   if (!renderInterior(p,bgndRect.adjusted(0,0,dw,dh),fspec,ispec,ispec.element+suffix))
   { // no window interior element but with reduced translucency
-    p->fillRect(bgndRect, QApplication::palette().color(suffix.contains("-inactive")
-                                                          ? QPalette::Inactive
-                                                          : QPalette::Active,
-                                                        QPalette::Window));
+    p->fillRect(bgndRect, standardPalette().color(suffix.contains("-inactive")
+                                                    ? QPalette::Inactive
+                                                    : QPalette::Active,
+                                                  QPalette::Window));
   }
   if (ro > 0)
     p->restore();
@@ -259,6 +261,133 @@ bool Style::eventFilter(QObject *o, QEvent *e)
           }
         }
       }
+      else if (QCommandLinkButton *cbtn = qobject_cast<QCommandLinkButton*>(o))
+      {
+        /* as in Qt -> qcommandlinkbutton.cpp -> QCommandLinkButton::paintEvent()
+           but with modifications for taking into account icon states and text colors */
+
+        QStylePainter p(cbtn);
+        p.save();
+
+        QStyleOptionButton option;
+        option.initFrom(cbtn);
+        option.features |= QStyleOptionButton::CommandLinkButton;
+        option.text = QString();
+        option.icon = QIcon(); // we draw the icon ourself
+        if (cbtn->isChecked())
+          option.state |= State_On;
+        else if (cbtn->isDown())
+          option.state |= State_Sunken;
+        else if (cbtn->underMouse())
+          option.state |= State_MouseOver;
+
+        /* panel */
+        p.drawControl(QStyle::CE_PushButton, option);
+
+        const int leftMargin = 7;
+        const int topMargin = 10;
+        const int rightMargin = 4;
+        const int bottomMargin = 10;
+
+        int vOffset = 0, hOffset = 0;
+        if (cbtn->isDown() && !cbtn->isChecked())
+        {
+          vOffset = pixelMetric(QStyle::PM_ButtonShiftVertical);
+          hOffset = pixelMetric(QStyle::PM_ButtonShiftHorizontal);
+        }
+
+        bool isInactive(isWidgetInactive(cbtn));
+        const label_spec lspec = getLabelSpec("PanelButtonCommand");
+
+        /* find the state and set the text color accordingly */
+        int state;
+        QPalette pPalette = option.palette;
+        QColor col;
+        if (!cbtn->isEnabled())
+          state = 0;
+        else if (cbtn->isChecked())
+        {
+          state = 4;
+          if (isInactive)
+            col = getFromRGBA(lspec.toggleInactiveColor);
+          if (!col.isValid())
+            col = getFromRGBA(lspec.toggleColor);
+        }
+        else if (cbtn->isDown())
+        {
+          state = 3;
+          if (isInactive)
+            col = getFromRGBA(lspec.pressInactiveColor);
+          if (!col.isValid())
+            col = getFromRGBA(lspec.pressColor);
+        }
+        else if (cbtn->underMouse())
+        {
+          state = 2;
+          if (isInactive)
+            col = getFromRGBA(lspec.focusInactiveColor);
+          if (!col.isValid())
+            col = getFromRGBA(lspec.focusColor);
+        }
+        else
+        {
+          state = 1;
+          if (isInactive)
+            col = getFromRGBA(lspec.normalInactiveColor);
+          if (!col.isValid())
+            col = getFromRGBA(lspec.normalColor);
+        }
+        if (col.isValid())
+        {
+          opacifyColor(col);
+          pPalette.setColor(QPalette::ButtonText, col);
+        }
+
+        /* icon */
+        if (!cbtn->icon().isNull())
+          p.drawPixmap(leftMargin + hOffset, topMargin + vOffset,
+                       getPixmapFromIcon(cbtn->icon(),
+                                         getIconMode(state, isInactive, lspec),
+                                         cbtn->isChecked() ? QIcon::On : QIcon::Off,
+                                         cbtn->iconSize()));
+
+        int textflags = Qt::TextShowMnemonic;
+        if (!styleHint(QStyle::SH_UnderlineShortcut, &option, cbtn))
+          textflags |= Qt::TextHideMnemonic;
+        textflags |= Qt::AlignTop; // (see Kvantum.cpp -> Style::drawItemText)
+
+        QFont titleFont = cbtn->font();
+        titleFont.setBold(true);
+        titleFont.setPointSizeF(9.0);
+
+        /* title */
+        p.setFont(titleFont);
+        int textOffset = cbtn->icon().actualSize(cbtn->iconSize()).width() + leftMargin + 6;
+        QRect titleRect = cbtn->rect().adjusted(textOffset, topMargin, -rightMargin, 0);
+        if (cbtn->description().isEmpty())
+        {
+          QFontMetrics fm(titleFont);
+          titleRect.setTop(titleRect.top()
+                           + qMax(0, (cbtn->icon().actualSize(cbtn->iconSize()).height()
+                           - fm.height()) / 2));
+        }
+        p.drawItemText(titleRect.translated(hOffset, vOffset),
+                       textflags, pPalette, cbtn->isEnabled(), cbtn->text(), QPalette::ButtonText);
+
+        /* description */
+        textflags |= Qt::TextWordWrap | Qt::ElideRight;
+        QFont descriptionFont = cbtn->font();
+        descriptionFont.setPointSizeF(9.0);
+        p.setFont(descriptionFont);
+        QFontMetrics fm(titleFont);
+        int descriptionOffset = topMargin + fm.height();
+        QRect descriptionRect = cbtn->rect().adjusted(textOffset, descriptionOffset,
+                                                      -rightMargin, -bottomMargin);
+        p.drawItemText(descriptionRect.translated(hOffset, vOffset), textflags,
+                       pPalette, cbtn->isEnabled(), cbtn->description(), QPalette::ButtonText);
+        p.restore();
+        return true; // don't let QCommandLinkButton::paintEvent() be called
+      }
     }
     break;
 
@@ -399,7 +528,8 @@ bool Style::eventFilter(QObject *o, QEvent *e)
       {
         if (!w->hasFocus())
           animationStartState_ = "normal";
-        else if (!animationStartState_.startsWith("c-toggled")) // the popup may have been closed (with Qt5)
+        /* the popup may have been closed (with Qt5) */
+        else if (!(animatedWidget_ == w && animationStartState_.startsWith("c-toggled")))
           animationStartState_ = "pressed";
         if (isWidgetInactive(w))
           animationStartState_.append("-inactive");
@@ -432,8 +562,9 @@ bool Style::eventFilter(QObject *o, QEvent *e)
           animationOpacity_ = 100;
           animatedWidget_->update();
         }
-        if (!animationStartState_.startsWith("c-toggled")
-            && !animationStartState_.startsWith("normal"))
+        if (!(animatedWidget_ == w
+              && (animationStartState_.startsWith("c-toggled")
+                  || animationStartState_.startsWith("normal"))))
         { // it was hidden or another widget was interacted with  -- there's no other possibility
           animationStartState_ = "normal";
           if (isWidgetInactive(w))
@@ -446,7 +577,9 @@ bool Style::eventFilter(QObject *o, QEvent *e)
       else
       {
         QAbstractScrollArea *sa = qobject_cast<QAbstractScrollArea*>(o);
-        if ((sa && !w->inherits("QComboBoxListView")) // exclude combo popups
+        if ((sa && !w->inherits("QComboBoxListView") // exclude combo popups
+             /* no animation without the top focused generic frame */
+             && elementExists(getFrameSpec(QStringLiteral("GenericFrame")).element+"-focused-top"))
             || (qobject_cast<QLineEdit*>(o)
                 // this is only needed for Qt5 -- Qt4 combo lineedits don't have FocusIn event
                 && !qobject_cast<QComboBox*>(w->parentWidget()))
@@ -506,7 +639,9 @@ bool Style::eventFilter(QObject *o, QEvent *e)
           || qobject_cast<QLineEdit*>(o)
           || qobject_cast<QAbstractSpinBox*>(o)
           || (qobject_cast<QAbstractScrollArea*>(o)
-              && !w->inherits("QComboBoxListView"))) // exclude combo popups
+              && !w->inherits("QComboBoxListView") // exclude combo popups
+              /* no animation without the top focused generic frame */
+              && elementExists(getFrameSpec(QStringLiteral("GenericFrame")).element+"-focused-top")))
       {
         /* disable animation if focus-out happens immediately after focus-in
            for exactly the same area to prevent flashing */
@@ -704,17 +839,19 @@ bool Style::eventFilter(QObject *o, QEvent *e)
     {
       QPalette palette = w->palette();
       if (palette.color(QPalette::Active, QPalette::Text)
-          != QApplication::palette().color(QPalette::Active, QPalette::Text))
+          != standardPalette().color(QPalette::Active, QPalette::Text))
       { // Custom text color; don't set palettes! The app is responsible for all colors.
         break;
       }
       const label_spec lspec = getLabelSpec(QStringLiteral("ItemView"));
       /* set the normal inactive text color to the normal active one
          (needed when the app sets it inactive) */
-      QColor normalCol = getFromRGBA(lspec.normalColor);
-      if (!normalCol.isValid())
-        normalCol = QApplication::palette().color(QPalette::Active,QPalette::Text);
-      palette.setColor(QPalette::Inactive, QPalette::Text, normalCol);
+      QColor col = getFromRGBA(lspec.normalColor);
+      if (!col.isValid())
+        col = standardPalette().color(QPalette::Active,QPalette::Text);
+      else
+        opacifyColor(col);
+      palette.setColor(QPalette::Inactive, QPalette::Text, col);
       if (!hasInactiveSelItemCol_)
       {
         w->setPalette(palette);
@@ -722,15 +859,16 @@ bool Style::eventFilter(QObject *o, QEvent *e)
       }
       /* set the toggled inactive text color to the toggled active one
          (the main purpose of installing an event filter on the view) */
-      palette.setColor(QPalette::Inactive, QPalette::HighlightedText,
-                       getFromRGBA(lspec.toggleColor));
+      col = getFromRGBA(lspec.toggleColor);
+      opacifyColor(col);
+      palette.setColor(QPalette::Inactive, QPalette::HighlightedText, col);
       /* use the active highlight color for the toggled (unfocused) item if there's
          no contrast with the pressed state because some apps (like Qt Designer)
          may not call PE_PanelItemViewItem but highlight the item instead */
       if (!toggledItemHasContrast_)
       {
         palette.setColor(QPalette::Inactive, QPalette::Highlight,
-                         QApplication::palette().color(QPalette::Active,QPalette::Highlight));
+                         standardPalette().color(QPalette::Active,QPalette::Highlight));
       }
       w->setPalette(palette);
     }
@@ -741,29 +879,32 @@ bool Style::eventFilter(QObject *o, QEvent *e)
     {
       QPalette palette = w->palette();
       if (palette.color(QPalette::Active, QPalette::Text)
-          != QApplication::palette().color(QPalette::Active, QPalette::Text))
+          != standardPalette().color(QPalette::Active, QPalette::Text))
       {
         break;
       }
       const label_spec lspec = getLabelSpec(QStringLiteral("ItemView"));
       /* restore the normal inactive text color (which was changed at QEvent::WindowActivate) */
-      QColor normalInactiveCol = getFromRGBA(lspec.normalInactiveColor);
-      if (!normalInactiveCol.isValid())
-        normalInactiveCol = QApplication::palette().color(QPalette::Inactive,QPalette::Text);
-      palette.setColor(QPalette::Inactive, QPalette::Text, normalInactiveCol);
+      QColor col = getFromRGBA(lspec.normalInactiveColor);
+      if (!col.isValid())
+        col = standardPalette().color(QPalette::Inactive,QPalette::Text);
+      else
+        opacifyColor(col);
+      palette.setColor(QPalette::Inactive, QPalette::Text, col);
       if (!hasInactiveSelItemCol_)
       { // custom text color
         w->setPalette(palette);
         break;
       }
       /* restore the toggled inactive text color (which was changed at QEvent::WindowActivate) */
-      palette.setColor(QPalette::Inactive,QPalette::HighlightedText,
-                       getFromRGBA(lspec.toggleInactiveColor));
+      col = getFromRGBA(lspec.toggleInactiveColor);
+      opacifyColor(col);
+      palette.setColor(QPalette::Inactive,QPalette::HighlightedText, col);
       /* restore the inactive highlight color (which was changed at QEvent::WindowActivate) */
       if (!toggledItemHasContrast_)
       {
         palette.setColor(QPalette::Inactive, QPalette::Highlight,
-                         QApplication::palette().color(QPalette::Inactive,QPalette::Highlight));
+                         standardPalette().color(QPalette::Inactive,QPalette::Highlight));
       }
       w->setPalette(palette);
     }
@@ -794,11 +935,18 @@ bool Style::eventFilter(QObject *o, QEvent *e)
       else if (qobject_cast<QMenu*>(o))
 #endif
       {
+        //if (isLibreoffice_) break;
+        if (w->testAttribute(Qt::WA_X11NetWmWindowTypeMenu)) break; // detached menu
+#if (QT_VERSION >= QT_VERSION_CHECK(5,12,0))
+        if (w->testAttribute(Qt::WA_StyleSheetTarget)) break; // not drawn by Kvantum (see PM_MenuHMargin)
+#endif
         if (movedMenus.contains(w)) break; // already moved
         /* "magical" condition for a submenu */
         QPoint parentMenuCorner;
         QMenu *parentMenu = qobject_cast<QMenu*>(QApplication::activePopupWidget());
-        if (!parentMenu)
+        /* this is a workaround for KDE global menu */
+        if (parentMenu && parentMenu->activeAction() == nullptr) parentMenu = nullptr;
+        if (parentMenu == nullptr)
         { // search for a detached menu with an active action
           const QWidgetList topLevels = QApplication::topLevelWidgets();
           for (QWidget *topWidget : topLevels)
@@ -817,7 +965,7 @@ bool Style::eventFilter(QObject *o, QEvent *e)
         else
           parentMenuCorner = parentMenu->mapToGlobal(QPoint(0,0));
         QMenuBar *parentMenubar = nullptr;
-        if (!parentMenu)
+        if (parentMenu == nullptr)
         { // search for a menubar with an active action
           if (QMainWindow *mw = qobject_cast<QMainWindow*>(QApplication::activeWindow()))
           {
@@ -830,17 +978,19 @@ bool Style::eventFilter(QObject *o, QEvent *e)
         }
         /* get the available geometry (Qt menus don't
            spread across the available virtual geometry) */
-        QRect ag;
+        QRect sg, ag;
         if (QWindow *win = w->windowHandle())
         {
           if (QScreen *sc = win->screen())
+          {
+            sg = sc->geometry();
             ag = sc->availableGeometry();
+          }
         }
+        qreal dX = 0, dY = 0;
         /* this gives the real position AFTER pending movements
            because it's QWidgetData::crect (Qt -> qwidget.h) */
         QRect g(w->geometry());
-        int X = g.left();
-        int Y = g.top();
 
         /* WARNING: If compositing is stopped here, we aren't responsible.
                     A check for the state of compositing at this very moment
@@ -850,19 +1000,19 @@ bool Style::eventFilter(QObject *o, QEvent *e)
         {
           /* compensate for the offset created by the shadow */
 
-          Y -= menuShadow_.at(1); // top shadow
+          dY -= menuShadow_.at(1); // top shadow
 
           if (w->layoutDirection() == Qt::RightToLeft)
           { // see explanations for ltr below
-            X += menuShadow_.at(2);
+            dX += menuShadow_.at(2);
             if (parentMenu)
             {
               if (parentMenuCorner.x() < g.left())
-                X -= menuShadow_.at(2) + menuShadow_.at(0);
+                dX -= menuShadow_.at(2) + menuShadow_.at(0);
               else
               {
-                X += menuShadow_.at(0)
-                     - getMenuMargin(true); // workaround for an old Qt bug
+                dX += menuShadow_.at(0)
+                      - static_cast<qreal>(getMenuMargin(true)); // workaround for an old Qt bug
               }
             }
             else
@@ -871,21 +1021,21 @@ bool Style::eventFilter(QObject *o, QEvent *e)
               {
                 QString group = tspec_.merge_menubar_with_toolbar ? "Toolbar" : "MenuBar";
                 if (parentMenubar->mapToGlobal(QPoint(0,0)).y() > g.bottom())
-                  Y +=  menuShadow_.at(1) + menuShadow_.at(3) + getFrameSpec(group).top;
+                  dY +=  menuShadow_.at(1) + menuShadow_.at(3) + static_cast<qreal>(getFrameSpec(group).top);
                 else
-                  Y -= getFrameSpec(group).bottom;
+                  dY -= static_cast<qreal>(getFrameSpec(group).bottom);
 
                 QRect activeG = parentMenubar->actionGeometry(parentMenubar->activeAction());
                 QPoint activeTopLeft = parentMenubar->mapToGlobal(activeG.topLeft());
                 if (g.right() + 1 > activeTopLeft.x() + activeG.width())
                 { // Qt positions the menu wrongly in this case but we don't add a workaround
-                  X -= menuShadow_.at(2);
-                  int delta = menuShadow_.at(2)
-                              - (g.right() + 1 - (activeTopLeft.x() + activeG.width()));
+                  dX -= menuShadow_.at(2);
+                  qreal delta = menuShadow_.at(2)
+                                - static_cast<qreal>(g.right() + 1 - (activeTopLeft.x() + activeG.width()));
                   if (delta > 0)
-                    X += delta;
+                    dX += delta;
                   else
-                    X -= qMin(menuShadow_.at(0), -delta);
+                    dX -= qMin(menuShadow_.at(0), -delta);
                 }
               }
               else
@@ -894,39 +1044,39 @@ bool Style::eventFilter(QObject *o, QEvent *e)
                 {
                   QPoint wTopLeft = sunkenButton_.data()->mapToGlobal(QPoint(0,0));
                   if (wTopLeft.y() >= g.bottom())
-                    Y +=  menuShadow_.at(1) + menuShadow_.at(3);
+                    dY +=  menuShadow_.at(1) + menuShadow_.at(3);
                   if (g.right() + 1 > wTopLeft.x() + sunkenButton_.data()->width())
                   {
-                    X -= menuShadow_.at(2);
-                    int delta = menuShadow_.at(2)
-                                - (g.right() + 1 - (wTopLeft.x() + sunkenButton_.data()->width()));
+                    dX -= menuShadow_.at(2);
+                    qreal delta = menuShadow_.at(2)
+                                  - static_cast<qreal>(g.right() + 1 - (wTopLeft.x() + sunkenButton_.data()->width()));
                     if (delta > 0)
-                      X += delta;
+                      dX += delta;
                     else
-                      X -= qMin(menuShadow_.at(0), -delta);
+                      dX -= qMin(menuShadow_.at(0), -delta);
                   }
                 }
                 else if (!ag.isEmpty())
                 {
                   if (g.bottom() == ag.bottom() && g.top() != ag.top())
-                    Y += menuShadow_.at(1) + menuShadow_.at(3);
+                    dY += menuShadow_.at(1) + menuShadow_.at(3);
                   if (g.left() == ag.left() && g.right() != ag.right())
-                    X -= menuShadow_.at(2) + menuShadow_.at(0);
+                    dX -= menuShadow_.at(2) + menuShadow_.at(0);
                 }
               }
             }
           }
           else // ltr
           {
-            X -= menuShadow_.at(0); // left shadow
+            dX -= menuShadow_.at(0); // left shadow
             if (parentMenu)
             {
               if (parentMenuCorner.x() > g.left())
               { // there wasn't enough space to the right of the parent
-                X += menuShadow_.at(0) + menuShadow_.at(2);
+                dX += menuShadow_.at(0) + menuShadow_.at(2);
               }
               else
-                X -= menuShadow_.at(2); // right shadow of the left menu
+                dX -= menuShadow_.at(2); // right shadow of the left menu
             }
             else
             {
@@ -934,21 +1084,21 @@ bool Style::eventFilter(QObject *o, QEvent *e)
               {
                 QString group = tspec_.merge_menubar_with_toolbar ? "Toolbar" : "MenuBar";
                 if (parentMenubar->mapToGlobal(QPoint(0,0)).y() > g.bottom()) // menu is above menubar
-                  Y +=  menuShadow_.at(1) + menuShadow_.at(3) + getFrameSpec(group).top;
+                  dY +=  menuShadow_.at(1) + menuShadow_.at(3) + static_cast<qreal>(getFrameSpec(group).top);
                 else
-                  Y -= getFrameSpec(group).bottom;
+                  dY -= static_cast<qreal>(getFrameSpec(group).bottom);
 
                 QPoint activeTopLeft = parentMenubar->mapToGlobal(parentMenubar->actionGeometry(
                                                                    parentMenubar->activeAction())
                                                                  .topLeft());
                 if (activeTopLeft.x() > g.left()) // because of the right screen border
                 {
-                  X += menuShadow_.at(0);
-                  int delta = menuShadow_.at(0) - (activeTopLeft.x() - g.left());
+                  dX += menuShadow_.at(0);
+                  qreal delta = menuShadow_.at(0) - static_cast<qreal>(activeTopLeft.x() - g.left());
                   if (delta > 0)
-                    X -= delta;
+                    dX -= delta;
                   else
-                    X += qMin(menuShadow_.at(2), -delta);
+                    dX += qMin(menuShadow_.at(2), -delta);
                 }
               }
               else
@@ -957,63 +1107,80 @@ bool Style::eventFilter(QObject *o, QEvent *e)
                 {
                   QPoint wTopLeft = sunkenButton_.data()->mapToGlobal(QPoint(0,0));
                   if (wTopLeft.y() >= g.bottom()) // above the button (strange! Qt doesn't add 1px)
-                    Y +=  menuShadow_.at(1) + menuShadow_.at(3);
+                    dY +=  menuShadow_.at(1) + menuShadow_.at(3);
                   if (wTopLeft.x() > g.left()) // because of the right screen border
                   {
-                    X += menuShadow_.at(0);
-                    int delta = menuShadow_.at(0) - (wTopLeft.x() - g.left());
+                    dX += menuShadow_.at(0);
+                    qreal delta = menuShadow_.at(0) - static_cast<qreal>(wTopLeft.x() - g.left());
                     if (delta > 0)
-                      X -= delta;
+                      dX -= delta;
                     else
-                      X += qMin(menuShadow_.at(2), -delta);
+                      dX += qMin(menuShadow_.at(2), -delta);
                   }
                 }
                 else if (!ag.isEmpty()) // probably a panel menu
                 {
                   /* snap to the screen bottom if possible */
                   if (g.bottom() == ag.bottom() && g.top() != ag.top())
-                    Y += menuShadow_.at(1) + menuShadow_.at(3);
+                    dY += menuShadow_.at(1) + menuShadow_.at(3);
                   /* snap to the right screen edge if possible */
                   if (g.right() == ag.right() && g.left() != ag.left())
-                    X += menuShadow_.at(0) + menuShadow_.at(2);
+                    dX += menuShadow_.at(0) + menuShadow_.at(2);
                 }
               }
             }
           }
-
-#if (QT_VERSION >= QT_VERSION_CHECK(5,11,0))
-          /* compensate for an annoyance in Qt 5.11 -> QMenu::internalDelayedPopup() */
-          if (parentMenu)
-          {
-            QAction *activeAct = parentMenu->activeAction();
-            if (activeAct && activeAct == menu->menuAction() && activeAct->isEnabled()
-                && activeAct->menu() && activeAct->menu()->isEnabled() && activeAct->menu()->isVisible())
-            {
-              const auto &actions = w->actions();
-              if (!actions.isEmpty())
-              {
-                const auto topActionRect = menu->actionGeometry(actions.first());
-                if (g.top() + topActionRect.top() ==  parentMenu->actionGeometry(activeAct).top()
-                                                      + parentMenuCorner.y())
-                {
-                  Y += topActionRect.top();
-                }
-              }
-            }
-          }
-#endif
-
-          w->move(X,Y);
         }
         else if (!parentMenu && parentMenubar)
         {
           QString group = tspec_.merge_menubar_with_toolbar ? "Toolbar" : "MenuBar";
           if (parentMenubar->mapToGlobal(QPoint(0,0)).y() > g.bottom()) // menu is above menubar
-            Y += getFrameSpec(group).top;
+            dY += static_cast<qreal>(getFrameSpec(group).top);
           else
-            Y -= getFrameSpec(group).bottom;
-          w->move(X,Y);
+            dY -= static_cast<qreal>(getFrameSpec(group).bottom);
         }
+
+#if (QT_VERSION >= QT_VERSION_CHECK(5,11,0))
+        /* compensate for an annoyance in Qt 5.11 -> QMenu::internalDelayedPopup() */
+        if (parentMenu)
+        {
+          QAction *activeAct = parentMenu->activeAction();
+          if (activeAct && activeAct == menu->menuAction() && activeAct->isEnabled()
+              && activeAct->menu() && activeAct->menu()->isEnabled() && activeAct->menu()->isVisible())
+          {
+            const auto &actions = w->actions();
+            if (!actions.isEmpty())
+            {
+              const auto topActionRect = menu->actionGeometry(actions.first());
+              if (g.top() + topActionRect.top() ==  parentMenu->actionGeometry(activeAct).top()
+                                                    + parentMenuCorner.y())
+              {
+                dY += static_cast<qreal>(topActionRect.top());
+              }
+            }
+          }
+        }
+#endif
+
+        int DX = qRound(dX);
+        int DY = qRound(dY);
+        if (DX == 0 && DY == 0) break;
+
+        // prevent the menu from switching to another screen
+        if (!sg.isEmpty() && QApplication::screens().size() > 1)
+        {
+          if (g.top() + DY < sg.top() && g.top() >= sg.top())
+            DY = sg.top() - g.top();
+          if (w->layoutDirection() == Qt::RightToLeft)
+          {
+            if (g.right() + DX > sg.right() && g.right() <= sg.right())
+              DX = sg.right() - g.right();
+          }
+          else if (g.left() + DX < sg.left() && g.left() >= sg.left())
+            DX = sg.left() - g.left();
+        }
+
+        w->move(g.left() + DX, g.top() + DY);
         movedMenus.insert(w);
       }
       else if (tspec_.group_toolbar_buttons && qobject_cast<QToolButton*>(o))
@@ -1031,47 +1198,53 @@ bool Style::eventFilter(QObject *o, QEvent *e)
            (-> QEvent::WindowActivate and QEvent::WindowDeactivate) */
         QPalette palette = w->palette();
         if (palette.color(QPalette::Active, QPalette::Text)
-            != QApplication::palette().color(QPalette::Active, QPalette::Text))
+            != standardPalette().color(QPalette::Active, QPalette::Text))
         {
           break;
         }
         const label_spec lspec = getLabelSpec(QStringLiteral("ItemView"));
         if (isWidgetInactive(w)) // FIXME: probably not needed with inactive window
         {
-          QColor normalInactiveCol = getFromRGBA(lspec.normalInactiveColor);
-          if (!normalInactiveCol.isValid())
-            normalInactiveCol = QApplication::palette().color(QPalette::Inactive,QPalette::Text);
-          palette.setColor(QPalette::Inactive, QPalette::Text, normalInactiveCol);
+          QColor col = getFromRGBA(lspec.normalInactiveColor);
+          if (!col.isValid())
+            col = standardPalette().color(QPalette::Inactive,QPalette::Text);
+          else
+            opacifyColor(col);
+          palette.setColor(QPalette::Inactive, QPalette::Text, col);
           if (!hasInactiveSelItemCol_)
           {
             w->setPalette(palette);
             break;
           }
-          palette.setColor(QPalette::Inactive,QPalette::HighlightedText,
-                           getFromRGBA(lspec.toggleInactiveColor));
+          col = getFromRGBA(lspec.toggleInactiveColor);
+          opacifyColor(col);
+          palette.setColor(QPalette::Inactive,QPalette::HighlightedText, col);
           if (!toggledItemHasContrast_)
           {
             palette.setColor(QPalette::Inactive, QPalette::Highlight,
-                             QApplication::palette().color(QPalette::Inactive,QPalette::Highlight));
+                             standardPalette().color(QPalette::Inactive,QPalette::Highlight));
           }
         }
         else
         {
-          QColor normalCol = getFromRGBA(lspec.normalColor);
-          if (!normalCol.isValid())
-            normalCol = QApplication::palette().color(QPalette::Active,QPalette::Text);
-          palette.setColor(QPalette::Inactive, QPalette::Text, normalCol);
+          QColor col = getFromRGBA(lspec.normalColor);
+          if (!col.isValid())
+            col = standardPalette().color(QPalette::Active,QPalette::Text);
+          else
+            opacifyColor(col);
+          palette.setColor(QPalette::Inactive, QPalette::Text, col);
           if (!hasInactiveSelItemCol_)
           {
             w->setPalette(palette);
             break;
           }
-          palette.setColor(QPalette::Inactive, QPalette::HighlightedText,
-                           getFromRGBA(lspec.toggleColor));
+          col = getFromRGBA(lspec.toggleColor);
+          opacifyColor(col);
+          palette.setColor(QPalette::Inactive, QPalette::HighlightedText, col);
           if (!toggledItemHasContrast_)
           {
             palette.setColor(QPalette::Inactive, QPalette::Highlight,
-                             QApplication::palette().color(QPalette::Active,QPalette::Highlight));
+                             standardPalette().color(QPalette::Active,QPalette::Highlight));
           }
         }
         w->setPalette(palette);
@@ -1088,7 +1261,7 @@ bool Style::eventFilter(QObject *o, QEvent *e)
   /* FIXME For some reason unknown to me (a Qt5 bug?), the Qt5 spinbox size hint
      is sometimes wrong as if Qt5 spinboxes don't have time to consult CT_SpinBox
      although they should (-> qabstractspinbox.cpp -> QAbstractSpinBox::sizeHint).
-     The same thing rarely happens with Qt4 too. Here we force a minimum size by
+     The same thing rarely happened with Qt4 too. Here we force a minimum size by
      using CT_SpinBox when the maximum size isn't set by the app or isn't smaller
      than our size. */
   case QEvent::ShowToParent:
@@ -1099,7 +1272,7 @@ bool Style::eventFilter(QObject *o, QEvent *e)
             || qobject_cast<QDoubleSpinBox*>(o)
             || qobject_cast<QDateTimeEdit*>(o)))
     {
-      QSize size = sizeFromContents(CT_SpinBox,nullptr,QSize(),w);
+      QSize size = sizeFromContents(CT_SpinBox,nullptr,QSize(),w).expandedTo(w->minimumSize());
       if (w->maximumWidth() > size.width())
         w->setMinimumWidth(size.width());
       if (w->maximumHeight() > size.height())
@@ -1111,18 +1284,33 @@ bool Style::eventFilter(QObject *o, QEvent *e)
                  || !getInteriorSpec(QStringLiteral("ToolbarLineEdit")).element.isEmpty())
              && getStylableToolbarContainer(w, true))
     {
-      QPalette palette = w->palette();
       const label_spec tlspec = getLabelSpec("Toolbar");
       QColor col = getFromRGBA(tlspec.normalColor);
-      if (col.isValid() && palette.color(QPalette::Active, QPalette::Text) != col)
+      if (enoughContrast(col, standardPalette().color(QPalette::Active,QPalette::Text)))
       {
-        palette.setColor(QPalette::Active, QPalette::Text, col);
-        QColor col1 = getFromRGBA(tlspec.normalInactiveColor);
+        QColor col1 = col;
+        opacifyColor(col1);
+        QPalette palette = w->palette();
+        palette.setColor(QPalette::Active, QPalette::Text, col1);
+#if (QT_VERSION >= QT_VERSION_CHECK(5,12,0))
+        col1 = col; // placeholder
+        col1.setAlpha(128);
+        opacifyColor(col1);
+        palette.setColor(QPalette::PlaceholderText, col1);
+#endif
+        col1 = getFromRGBA(tlspec.normalInactiveColor);
         if (!col1.isValid()) col1 = col;
+        opacifyColor(col1);
         palette.setColor(QPalette::Inactive, QPalette::Text, col1);
         col.setAlpha(102); // 0.4 * disabledCol.alpha()
+        opacifyColor(col);
         palette.setColor(QPalette::Disabled, QPalette::Text,col);
         w->setPalette(palette);
+#if (QT_VERSION >= QT_VERSION_CHECK(5,12,0))
+        /* also correct the color of the symbolic clear icon (-> CE_ToolBar) */
+        if (QAction *clearAction = w->findChild<QAction*>(QLatin1String("_q_qlineeditclearaction")))
+          clearAction->setIcon(standardIcon(QStyle::SP_LineEditClearButton, nullptr, w));
+#endif
       }
     }
     break;
